/** 
 * File: ChessClock01.ino
 * Description:
 * Software for Arduino driven chess clock with solenoid button-release mechanism.
 * For details wisit: https://doc.csipa.hu/?p=254
 * Known bugs:
 *   - dim stays when exiting finish mode
 * 
 * Author: Balazs Kelemen
 * Contact: prampec+arduino@gmail.com
 * Copyright: 2017 Balazs Kelemen
 * Copying permission statement:
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

#include <PciManager.h>
#include <Debouncer.h>
#include <SoftTimer.h>
#include <Rotary.h>
#include <TM1637Display.h>

// -- PINS
#define ROT_PIN_A A1
#define ROT_PIN_B A0
#define ROT_PUSH A2

#define SWITCH_L 12
#define SOLENOID_L 13
#define SWITCH_R 2
#define SOLENOID_R 3

#define DISP_CLK_L 10
#define DISP_DIO_L 11
#define DISP_CLK_R 5
#define DISP_DIO_R 4

#define BAT A3

void onRotate(short direction, Rotary* rotary);
void onRotPushPress();
void onRotPushRelease(unsigned long pressTime);
void onSwitchLeftPress();
void onSwitchRightPress();
boolean solenoidsOff(Task* task);
void checkBothPushed(Task* task);
void displayChessStateCallback(Task* task);
void fickerLooseSideCallback(Task* task);

Rotary r(ROT_PIN_A, ROT_PIN_B, onRotate, true);
Debouncer rotPushDebouncer(ROT_PUSH, MODE_CLOSE_ON_PUSH, onRotPushPress, onRotPushRelease, true);
Debouncer swLeftDebouncer(SWITCH_L, MODE_CLOSE_ON_PUSH, onSwitchLeftPress, NULL, true);
Debouncer swRightDebouncer(SWITCH_R, MODE_CLOSE_ON_PUSH, onSwitchRightPress, NULL, true);
TM1637Display displayLeft(DISP_CLK_L, DISP_DIO_L);
TM1637Display displayRight(DISP_CLK_R, DISP_DIO_R);
DelayRun solenoidsOffTask(100, solenoidsOff);
Task checkBothPushedTask(100, checkBothPushed);
Task displayChessStateTask(0, displayChessStateCallback);
Task fickerLooseSide(1000, fickerLooseSideCallback);

#define LONG_PRESS_MS 1500

#define LEFT 0
#define RIGHT 1

#define STATE_CHESS_SETUP 20
#define STATE_CHESS_RUN   21
#define STATE_CHESS_FINISHED 22

byte state = STATE_CHESS_SETUP;

unsigned long chessTimeSecs = 600;
byte activeSide = LEFT;
unsigned long timePassedForLeftMs;
unsigned long timePassedForRightMs;
unsigned long lastChangeMs;
TM1637Display* looseSide;

void setup() {
  Serial.begin(9600);
  PciManager.registerListener(ROT_PUSH, &rotPushDebouncer);
  PciManager.registerListener(SWITCH_L, &swLeftDebouncer);
  PciManager.registerListener(SWITCH_R, &swRightDebouncer);

  pinMode(SOLENOID_L, OUTPUT);
  pinMode(SOLENOID_R, OUTPUT);

  activeSide = digitalRead(SWITCH_L) ? LEFT : RIGHT;

  SoftTimer.add(&checkBothPushedTask);

  displayLeft.setBrightness(0x0f);
  displayRight.setBrightness(0x0f);
  uint8_t data[] = { SEG_G, SEG_G, SEG_G, SEG_G };
  displayLeft.setSegments(data);
  displayRight.setSegments(data);

  setState(STATE_CHESS_SETUP);

  Serial.println("Ready.");
}

void setState(byte newState) {
  state = newState;
  if (state == STATE_CHESS_SETUP) {
    SoftTimer.remove(&displayChessStateTask);
    SoftTimer.remove(&fickerLooseSide);
    displayChessTime();
  }
  else if (state == STATE_CHESS_RUN)
  {
    displayTime(&displayLeft, chessTimeSecs, true);
    displayTime(&displayRight, chessTimeSecs, true);
    SoftTimer.add(&displayChessStateTask);
    timePassedForLeftMs = 0;
    timePassedForRightMs = 0;
    lastChangeMs = millis();
  }
  else if (state == STATE_CHESS_FINISHED)
  {
    SoftTimer.remove(&displayChessStateTask);
    SoftTimer.add(&fickerLooseSide);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////

void displayChessTime() {
  uint8_t data[] = { 
    SEG_A | SEG_D | SEG_E | SEG_F,
    SEG_B | SEG_C | SEG_E | SEG_F | SEG_G,
    SEG_A | SEG_D | SEG_E | SEG_F | SEG_G,
    0
    };
  displayLeft.setSegments(data);  
  displayTime(&displayRight, chessTimeSecs, true);
  Serial.print("Displaying ");
  Serial.println(chessTimeSecs);
}

unsigned long lastDisplayedTime = 0;
boolean lastDisplayedDots = true;
void displayChessState() {
  TM1637Display* display = activeSide == LEFT ? &displayLeft : &displayRight;
  unsigned long timeMs = getActualTimePassedForSide(activeSide);
  unsigned long timeLeft = chessTimeSecs - timeMs / 1000;
  boolean displayDots = (timeMs / 250) % 2;
  if (timeLeft <= 0)
  {
    setState(STATE_CHESS_FINISHED);
    timeLeft = 0;
    looseSide = display;
  }
  if ((timeLeft != lastDisplayedTime) || (lastDisplayedDots != displayDots))
  {
    displayTime(display, timeLeft, displayDots);
    lastDisplayedTime = timeLeft;
    lastDisplayedDots = displayDots;
    Serial.print("Displaying ");
    Serial.println(timeLeft);
  }
}

//////////////////////////////////////////////////////////////////////////////////////////

void onRotate(short direction, Rotary* rotary) {
  if(direction == DIRECTION_CW) {
    Serial.println("cw");
  }
  if(direction == DIRECTION_CCW) {
    Serial.println("ccw");
  }
  if (state == STATE_CHESS_SETUP) {
    if ((direction > 0) || (chessTimeSecs > 60)) {
      chessTimeSecs += direction * 60;
      displayChessTime();
    }
  }
}

void onRotPushPress() {
  Serial.println("pushed");
  if (state == STATE_CHESS_SETUP)
  {
    setState(STATE_CHESS_RUN);
  }
}
void onRotPushRelease(unsigned long pressTime) {
  Serial.println("released");
  if (((state == STATE_CHESS_RUN) || (state == STATE_CHESS_FINISHED)) && (pressTime >= LONG_PRESS_MS)) {
    setState(STATE_CHESS_SETUP);
  }
}
void onSwitchLeftPress() {
  Serial.println("left");
  updateActualTimePassed();
  digitalWrite(SOLENOID_R, HIGH);
  solenoidsOffTask.startDelayed();
  Serial.println("Solenoid R On");
  activeSide = RIGHT;
}
void onSwitchRightPress() {
  updateActualTimePassed();
  Serial.println("right");
  digitalWrite(SOLENOID_L, HIGH);
  solenoidsOffTask.startDelayed();
  Serial.println("Solenoid L On");
  activeSide = LEFT;
}

//////////////////////////////////////////////////////////////////////////////////////////

boolean solenoidsOff(Task* task) {
  digitalWrite(SOLENOID_R, LOW);
  digitalWrite(SOLENOID_L, LOW);
  Serial.println("Solenoids Off");
  return true;
}

void checkBothPushed(Task* task) {
  // TODO: delegate this to debouncer
  /*
  if ((!digitalRead(SOLENOID_R)) && (!digitalRead(SOLENOID_L)))
  {
    Serial.println("Both are pushed");
    int sidePin = activeSide == LEFT ? SOLENOID_L : SOLENOID_R;
    digitalWrite(sidePin, HIGH);
    solenoidsOffTask.startDelayed();
  }
  */
}

void displayChessStateCallback(Task* task) {
  displayChessState();
}

boolean dim = true;
void fickerLooseSideCallback(Task* task) {
  looseSide->setBrightness(dim ? 0x08 : 0x0f);
  displayTime(looseSide, 0, true);
  dim = 1 - dim;
}

//////////////////////////////////////////////////////////////////////////////////////////

void displayTime(TM1637Display* display, unsigned long time, boolean dot)
{
  unsigned long mins = time / 60;
  unsigned long secs = time % 60;
  unsigned long timeNumber = mins * 100 + secs;
  displayTimeNumber(display, timeNumber, dot);
}
void displayTimeNumber(TM1637Display* display, unsigned long number, boolean dot)
{
  uint8_t digits[] = { 0, 0, 0, 0 };
  digits[3] = display->encodeDigit(number % 10);
  number /= 10;
  digits[2] = display->encodeDigit(number % 10);
  number /= 10;
  uint8_t d = display->encodeDigit(number % 10);
  if (dot)
  {
    d = addDot(d);
  }
  digits[1] = d;
  number /= 10;
  int m = number % 10;
  if (m > 0) {
    digits[0] = display->encodeDigit(m);
  }
  display->setSegments(digits, 4, 0);
}

uint8_t addDot(uint8_t digit)
{
  return digit | 0x80;
}

unsigned long getActualTimePassedForSide(byte side)
{
  unsigned long value = side == LEFT ? timePassedForLeftMs : timePassedForRightMs;
  if (side == activeSide)
  {
    value += millis() - lastChangeMs; // -- TODO: handle millis overflow!
  }
  return value;
}

void updateActualTimePassed()
{
  if (state == STATE_CHESS_RUN)
  {
    unsigned long now = millis();
    unsigned long value = now - lastChangeMs; // -- TODO: handle millis overflow!
    Serial.print("Decreasing with:");
    Serial.println(value);
    if (activeSide == LEFT)
    {
      timePassedForLeftMs += value;
    }
    else
    {
      timePassedForRightMs += value;
    }
    lastChangeMs = now;
    Serial.print("Last change time ms:");
    Serial.println(lastChangeMs);
    lastDisplayedTime = chessTimeSecs + 1;
  }
}
